use std::io;

pub mod reader;
pub mod writer;

pub use reader::TextReader;
pub use writer::TextWriter;

use crate::BinarySource;
use crate::BytesBinarySource;
use crate::Domain;
use crate::DomainDecode;
use crate::IOValue;
use crate::PlainValue;
use crate::Reader;

pub fn from_str<'de, D: Domain + 'static, Dec: DomainDecode<D>>(
    s: &'de str,
    decode_embedded: &mut Dec,
) -> io::Result<PlainValue<'static, D>> {
    BytesBinarySource::new(s.as_bytes()).text().next(false, decode_embedded)
}

pub fn iovalue_from_str(s: &str) -> io::Result<IOValue> {
    BytesBinarySource::new(s.as_bytes()).text().next_iovalue(false)
}

pub fn annotated_from_str<'de, D: Domain + 'static, Dec: DomainDecode<D>>(
    s: &'de str,
    decode_embedded: &mut Dec,
) -> io::Result<PlainValue<'static, D>> {
    BytesBinarySource::new(s.as_bytes()).text().next(true, decode_embedded)
}

pub fn annotated_iovalue_from_str(s: &str) -> io::Result<IOValue> {
    BytesBinarySource::new(s.as_bytes()).text().next_iovalue(true)
}

#[cfg(test)]
mod formatting_tests {
    use crate::IOValue;
    use crate::test_domain::Dom;
    use crate::ArcValue;

    #[test] fn format_debug_and_parse() {
        let v = "[1, {z: 2, a: #!\"One\"}, 3]".parse::<ArcValue<Dom>>().unwrap();
        assert_eq!(format!("{:?}", &v), "[1, {a: #!\"One\", z: 2}, 3]");
    }

    #[test] fn format_pretty_debug_and_parse() {
        let v = "[1, {z: 2, a: #!\"One\"}, 3]".parse::<ArcValue<Dom>>().unwrap();
        assert_eq!(format!("{:#?}", &v), concat!(
            "[\n",
            "    1,\n",
            "    {\n",
            "        a: #!\"One\",\n",
            "        z: 2\n",
            "    },\n",
            "    3\n",
            "]"));
    }

    #[test] fn iovalue_parse() {
        let v = "[1 @{a:b c:d} @\"foo\" #![2 3] 4]".parse::<IOValue>().unwrap();
        assert_eq!(format!("{:#?}", &v), concat!(
            "[\n",
            "    1,\n",
            "    @{\n",
            "        a: b,\n",
            "        c: d\n",
            "    } @\"foo\" #![\n",
            "        2,\n",
            "        3\n",
            "    ],\n",
            "    4\n",
            "]"));
    }
}
